
from urllib import quote, unquote
from dircache import listdir
from os.path import join as pjoin
import errno, stat, os, sys

filename_suffix = ",w"
filename_suffix_len = len( filename_suffix )

class FileSystem:

	def __init__( s, basedir ):
		s.basedir =	basedir
		s.fcache =	{}

	def __getitem__( s, key ):
		try: return s.fcache[key]
		except KeyError:
			s.fcache[key] = File( s.basedir, key )
			return s.fcache[key]

	def has_key( s, key ):
		return ( not not s[key] )

	def keys( s ):
		return [ unquote( f[:-filename_suffix_len] )
			 for d1 in listdir( s.basedir )
			 for d2 in listdir( pjoin( s.basedir, d1 ))
			 for d3 in listdir( pjoin( s.basedir, d1, d2 ))
			 for f in listdir( pjoin( s.basedir, d1, d2, d3 ))
			 if f[-filename_suffix_len:] == filename_suffix ]


def writefile( name, contents ):
	f = open( name, 'wt' )
	try: f.write( contents )
	finally: f.close()


class File:

	prev_suffix = ",prev"

	def __init__( s, basedir, fkey ):
		s.basedir =	basedir
		s.fkey =	fkey
		s._contents =	None
		s.fname =	quote( s.fkey, '' ) + filename_suffix
		s.namehash =	"%010d" % abs( hash( s.fkey ))
		s.fullname =	pjoin( s.basedir, s.namehash[0:2],
					s.namehash[2:4], s.namehash[4:6],
					s.fname )

	def __len__( s ):
		if s._contents != None: return len(s._contents)
		try: return int( os.stat( s.fullname )[stat.ST_SIZE] )
		except OSError, e:
			if e.errno == errno.ENOENT: return 0
			else: raise

	def __nonzero__( s ):
		return len( s )

	def exists( s ):
		return len( s )

	def contents( s ):
		if not s.exists(): return ""
		if s._contents is None:
			s._contents = open( s.fullname, 'rt' ).read()
		return s._contents

	def file_diffs( s ):
		curr = s.fullname
		old = s.fullname + s.prev_suffix
		try: os.stat( old )
		except OSError, e:
			if e.errno == errno.ENOENT: old = '/dev/null'
			else: raise
		diff = os.popen( "diff -u %s %s" % ( old, curr ), 'rt' )
		return diff.readlines()

	def modtime( s ):
		try: return os.stat( s.fullname )[stat.ST_MTIME]
		except OSError, e:
			if e.errno == errno.ENOENT: return None
			else: raise

	def createdirs( s ):
		try: os.makedirs( os.path.dirname( s.fullname ))
		except OSError, e:
			if e.errno != errno.EEXIST: raise

	def tempname( s, suffix ):
		return "%s_%s_%d" % ( s.fullname, suffix, os.getpid() )

	def write( s, contents ):
		if contents == s.contents(): return
		s.createdirs()
		tempname = s.tempname( 'tmp' )
		savename = s.fullname + s.prev_suffix
		writefile( tempname, contents )
		try: os.unlink( savename )
		except OSError, e:
			if e.errno != errno.ENOENT: raise
		try: os.link( s.fullname, savename )
		except OSError, e:
			if e.errno != errno.ENOENT: raise
		os.rename( tempname, s.fullname )
		s._contents = None

	def mergechanges( s, contents, oldcontents ):
		# Warning! Race conditions abound.
		# I won't fix this until I get to know how it should be done.
		if contents == oldcontents: return None
		if not s.exists():
			s.write( contents )
			return False
		if contents == s.contents(): return None
		s.createdirs()
		yours = s.tempname( 'yours' )
		old = s.tempname( 'old' )
		writefile( yours, contents )
		writefile( old, oldcontents )
		mf = os.popen( "diff3 -Em -L mine -L old -L yours %s %s %s"
				% ( s.fullname, old, yours ), "r" )
		newcontents = mf.read()
		retval = not not mf.close()
		s.write( newcontents )
		os.unlink( yours )
		os.unlink( old )
		return retval

